#!/usr/bin/env python
# -*- coding: utf-8 -*-
# 
# Copyright 2016 The Aerospace Corporation.
# 
# This is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
# 
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this software; see the file COPYING.  If not, write to
# the Free Software Foundation, Inc., 51 Franklin Street,
# Boston, MA 02110-1301, USA.
# 
##################################################
# Message passing wrapper
#
# Makes block callbacks available through message
# passing service.
##################################################

from gnuradio import gr, blocks
import pmt
import inspect
import weakref
import json

# block wrapper
class message_passing_control_wrapper(gr.hier_block2):

    # nested class to forward messages to parameter setting methods
    class message_to_set_param(gr.sync_block):
        def __init__(self, setters, blklabel=None, ignore_error=True):
            #super(self.__class__, self)
            gr.sync_block.__init__(self, "message_to_set_param", None, None)
            self.setters = setters
            self.blklabel = blklabel
            self.ignore_error = ignore_error  # quietly discard sets of nonexistant parameters;
                                              # allows connection of one 'setter' block to
                                              # many 'settees' as long as they don't have the
                                              # same parameters, w/o logging many errors

            # get loggers
            alias = self.to_basic_block().alias() # not sure why alias is not inherited
            loggers = [x for x in gr.logger_get_names() if alias in x]
            gr.log.log_assert(len(loggers) == 2, "Too many loggers for %s" % alias)
            self.log, self.log_debug = tuple(gr.logger(l) for l in sorted(loggers))
            #self.log_debug.set_level('info')

            self.set_port = sport = "set_port"
            self.message_port_register_in(pmt.intern(sport))
            self.set_msg_handler(pmt.intern(sport), self.call_setter)

            self.print_port = pport = "print_port"
            self.message_port_register_in(pmt.intern(pport))
            self.set_msg_handler(pmt.intern(pport), self.print_msg)

        def print_msg(self, msg):
            # TODO: how to do this right using loggers, how to set the level and format?
            # TODO: how to integrate this into qa code?
            #import json
            # NOTE: gr_logger uses 1 argument to logger calls, different than python logging
            #self.log_debug.info("RECEIVED: %s" % pmt.write_string(msg))
            print "RECEIVED:", pmt.write_string(msg)
            #print json.dumps(pmt.to_python(msg))

        def call_setter(self, msg):
            if not pmt.is_dict(msg):
                self.log.error("Set param message must be a pmt dict of "
                               "{<param0>: [<arg0>, ...], ...}")
                return
            d = pmt.to_python(msg)
            #self.log_debug.debug("Received message %s" % pmt.write_string(msg))
            #self.log_debug.debug("Converted message %s" % json.dumps(d))
            for p, args in d.iteritems():
                try:
                    #self.log_debug.info('Calling %s.%s(%s)' % (self.blklabel,
                    #                                           self.setters[p].im_func.func_name,
                    #                                           ','.join('{0}'.format(x) for x in args)))
                    #from IPython import embed; embed()
                    self.setters[p](*args)
                # catch if args not a list or setters[p] not found
                except (KeyError, TypeError) as e:
                    self.log.errorIF(not self.ignore_error, str(e))

    ###### END message_to_set_param END #######

    def __init__(self, klass, *args, **kwds):
        self.blk = blk = klass(*args, **kwds)
        super(message_passing_control_wrapper, self).__init__(
            "m2sp_%s" % blk.name(),
            self.fix_signature(blk.input_signature()),
            self.fix_signature(blk.output_signature()))

        # connect inputs and outputs
        for n in xrange(self.input_signature().max_streams()):
            self.connect((self, n), (blk, n))
        for n in xrange(self.output_signature().max_streams()):
            self.connect((blk, n), (self, n))

        # map methods to method names
        methods = {n: m for n, m in inspect.getmembers(blk, inspect.ismethod)}
        # parameters getter/setter naming convention is usually <param>(), set_<param>()
        setters = {n:methods["set_%s" % n] for n in methods if "set_%s" % n in methods}

        self.proxy_blk = self.message_to_set_param(setters, blk.alias())

        # create and connect set_param port
        self.set_port = set_port = "set"
        # NOTE: the next line used to be self.message_port_register_hier_out(...)
        # check https://lists.gnu.org/archive/html/discuss-gnuradio/2016-01/msg00019.html
        # for details
        self.message_port_register_hier_in(set_port)
        self.msg_connect(weakref.proxy(self), set_port,
                         self.proxy_blk, self.proxy_blk.set_port)

    # TODO: add parameters to further modify signature?
    def fix_signature(self, sig):
        """Return input/output signatures with fixed port counts."""
        fixedsig = sig
        if sig.max_streams() < 0:  # hier_block2 cannot handle variable ports
            stream_sizes = sig.sizeof_stream_items()       # actual number of inputs from sizes vector
            fixedsig = gr.io_signature(sig.min_streams(), len(stream_sizes),
                                           stream_sizes[0] if len(stream_sizes) < 2
                                           else stream_sizes)
        return fixedsig

    def get_control_params(self):
        return self.proxy_blk.setters.keys()
